"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.compare = void 0;
const is_subset_of_1 = require("is-subset-of");
const typedescriptor_1 = require("typedescriptor");
const unbox_1 = require("./unbox");
const compare = {
    equal(left, right) {
        if (typedescriptor_1.Type.isArray(left) && typedescriptor_1.Type.isArray(right)) {
            if (left.length !== right.length) {
                return false;
            }
            for (const [i, leftItem] of left.entries()) {
                const rightItem = right[i];
                const areEqual = compare.equal(leftItem, rightItem);
                if (!areEqual) {
                    return false;
                }
            }
            return true;
        }
        if (typedescriptor_1.Type.isObject(left) && typedescriptor_1.Type.isObject(right)) {
            const areEqual = is_subset_of_1.isSubsetOf(left, right) && is_subset_of_1.isSubsetOf(right, left);
            return areEqual;
        }
        if (typedescriptor_1.Type.isFunction(left) && typedescriptor_1.Type.isFunction(right)) {
            const leftSource = left.toString(), rightSource = right.toString();
            return leftSource === rightSource;
        }
        const leftUnboxed = unbox_1.unbox(left), rightUnboxed = unbox_1.unbox(right);
        return leftUnboxed === rightUnboxed;
    },
    notEqual(left, right) {
        return !compare.equal(left, right);
    },
    lessThan(left, right) {
        const leftType = typedescriptor_1.Type.of(left), rightType = typedescriptor_1.Type.of(right);
        if ((leftType !== rightType) ||
            (typedescriptor_1.Type.isFunction(left) && typedescriptor_1.Type.isFunction(right))) {
            return false;
        }
        if ((typedescriptor_1.Type.isArray(left) && typedescriptor_1.Type.isArray(right)) ||
            (typedescriptor_1.Type.isObject(left) && typedescriptor_1.Type.isObject(right))) {
            const isLessThan = is_subset_of_1.isSubsetOf(left, right) && !is_subset_of_1.isSubsetOf(right, left);
            return isLessThan;
        }
        const leftUnboxed = unbox_1.unbox(left), rightUnboxed = unbox_1.unbox(right);
        return leftUnboxed < rightUnboxed;
    },
    lessThanOrEqual(left, right) {
        return compare.lessThan(left, right) || compare.equal(left, right);
    },
    greaterThan(left, right) {
        return compare.lessThan(right, left);
    },
    greaterThanOrEqual(left, right) {
        return compare.lessThan(right, left) || compare.equal(right, left);
    },
    identity(left, right) {
        return left === right;
    },
    equalByStructure(left, right) {
        const leftType = typedescriptor_1.Type.of(left), rightType = typedescriptor_1.Type.of(right);
        if (leftType !== 'object' || rightType !== 'object') {
            return false;
        }
        const areEqualByStructure = is_subset_of_1.isSubsetOf.structural(left, right) &&
            is_subset_of_1.isSubsetOf.structural(right, left);
        return areEqualByStructure;
    },
    notEqualByStructure(left, right) {
        const leftType = typedescriptor_1.Type.of(left), rightType = typedescriptor_1.Type.of(right);
        if (leftType !== 'object' || rightType !== 'object') {
            return false;
        }
        return !compare.equalByStructure(left, right);
    },
    lessThanByStructure(left, right) {
        const leftType = typedescriptor_1.Type.of(left), rightType = typedescriptor_1.Type.of(right);
        if (leftType !== 'object' || rightType !== 'object') {
            return false;
        }
        const isLessThanByStructure = is_subset_of_1.isSubsetOf.structural(left, right) &&
            !is_subset_of_1.isSubsetOf.structural(right, left);
        return isLessThanByStructure;
    },
    lessThanOrEqualByStructure(left, right) {
        const leftType = typedescriptor_1.Type.of(left), rightType = typedescriptor_1.Type.of(right);
        if (leftType !== 'object' || rightType !== 'object') {
            return false;
        }
        return compare.lessThanByStructure(left, right) || compare.equalByStructure(left, right);
    },
    greaterThanByStructure(left, right) {
        const leftType = typedescriptor_1.Type.of(left), rightType = typedescriptor_1.Type.of(right);
        if (leftType !== 'object' || rightType !== 'object') {
            return false;
        }
        return compare.lessThanByStructure(right, left);
    },
    greaterThanOrEqualByStructure(left, right) {
        const leftType = typedescriptor_1.Type.of(left), rightType = typedescriptor_1.Type.of(right);
        if (leftType !== 'object' || rightType !== 'object') {
            return false;
        }
        return compare.lessThanByStructure(right, left) || compare.equalByStructure(right, left);
    },
    eq(left, right) {
        return compare.equal(left, right);
    },
    ne(left, right) {
        return compare.notEqual(left, right);
    },
    lt(left, right) {
        return compare.lessThan(left, right);
    },
    lte(left, right) {
        return compare.lessThanOrEqual(left, right);
    },
    gt(left, right) {
        return compare.greaterThan(left, right);
    },
    gte(left, right) {
        return compare.greaterThanOrEqual(left, right);
    },
    id(left, right) {
        return compare.identity(left, right);
    },
    eqs(left, right) {
        return compare.equalByStructure(left, right);
    },
    nes(left, right) {
        return compare.notEqualByStructure(left, right);
    },
    lts(left, right) {
        return compare.lessThanByStructure(left, right);
    },
    ltes(left, right) {
        return compare.lessThanOrEqualByStructure(left, right);
    },
    gts(left, right) {
        return compare.greaterThanByStructure(left, right);
    },
    gtes(left, right) {
        return compare.greaterThanOrEqualByStructure(left, right);
    }
};
exports.compare = compare;
